from __future__ import annotations
from glob import glob
import os
import shlex
import typing as t

import click
import pkg_resources

from tutor import hooks

from .__about__ import __version__

HERE = os.path.abspath(os.path.dirname(__file__))


hooks.Filters.CONFIG_UNIQUE.add_items(
    [
        ("CAIRN_CLICKHOUSE_PASSWORD", "{{ 20|random_string }}"),
        ("CAIRN_POSTGRESQL_PASSWORD", "{{ 20|random_string }}"),
        ("CAIRN_SUPERSET_SECRET_KEY", "{{ 20|random_string }}"),
        ("CAIRN_SSO_CLIENT_SECRET", "{{ 20|random_string }}"),
    ]
)
hooks.Filters.CONFIG_DEFAULTS.add_items(
    [
        ("CAIRN_VERSION", __version__),
        ("CAIRN_DOCKER_HOST_SOCK_PATH", "/var/run/docker.sock"),
        ("CAIRN_HOST", "data.{{ LMS_HOST }}"),
        # Clickhouse
        ("CAIRN_RUN_CLICKHOUSE", True),
        (
            "CAIRN_CLICKHOUSE_DOCKER_IMAGE",
            "{{ DOCKER_REGISTRY }}overhangio/cairn-clickhouse:{{ CAIRN_VERSION }}",
        ),
        ("CAIRN_CLICKHOUSE_HOST", "cairn-clickhouse"),
        ("CAIRN_CLICKHOUSE_HTTP_PORT", 8123),
        ("CAIRN_CLICKHOUSE_HTTP_SCHEME", "http"),
        ("CAIRN_CLICKHOUSE_PORT", 9000),
        ("CAIRN_CLICKHOUSE_DATABASE", "openedx"),
        ("CAIRN_CLICKHOUSE_USERNAME", "openedx"),
        # Superset/Postgresql
        ("CAIRN_RUN_POSTGRESQL", True),
        ("CAIRN_POSTGRESQL_DATABASE", "superset"),
        ("CAIRN_POSTGRESQL_USERNAME", "superset"),
        (
            "CAIRN_SUPERSET_DOCKER_IMAGE",
            "{{ DOCKER_REGISTRY }}overhangio/cairn-superset:{{ CAIRN_VERSION }}",
        ),
        ("CAIRN_SUPERSET_LANGUAGE_CODE", "{{ LANGUAGE_CODE[:2] }}"),
        # SSO
        ("CAIRN_ENABLE_SSO", True),
        ("CAIRN_SSO_CLIENT_ID", "cairn"),
        # Vector
        # https://hub.docker.com/r/timberio/vector/tags
        # https://github.com/vectordotdev/vector/releases
        ("CAIRN_VECTOR_DOCKER_IMAGE", "docker.io/timberio/vector:0.25.1-alpine"),
    ]
)

# Init scripts
for service in ["cairn-clickhouse", "cairn-superset", "cairn-openedx"]:
    with open(
        os.path.join(HERE, "templates", "cairn", "tasks", service, "init"),
        encoding="utf-8",
    ) as fi:
        task = fi.read()
    hooks.Filters.CLI_DO_INIT_TASKS.add_item((service, task))

# Docker images
hooks.Filters.IMAGES_BUILD.add_items(
    [
        (
            "cairn-clickhouse",
            ("plugins", "cairn", "build", "cairn-clickhouse"),
            "{{ CAIRN_CLICKHOUSE_DOCKER_IMAGE }}",
            (),
        ),
        (
            "cairn-superset",
            ("plugins", "cairn", "build", "cairn-superset"),
            "{{ CAIRN_SUPERSET_DOCKER_IMAGE }}",
            (),
        ),
    ]
)
hooks.Filters.IMAGES_PULL.add_items(
    [
        (
            "cairn-clickhouse",
            "{{ CAIRN_CLICKHOUSE_DOCKER_IMAGE }}",
        ),
        (
            "cairn-superset",
            "{{ CAIRN_SUPERSET_DOCKER_IMAGE }}",
        ),
    ]
)
hooks.Filters.IMAGES_PUSH.add_items(
    [
        (
            "cairn-superset",
            "{{ CAIRN_SUPERSET_DOCKER_IMAGE }}",
        ),
        (
            "cairn-clickhouse",
            "{{ CAIRN_CLICKHOUSE_DOCKER_IMAGE }}",
        ),
    ]
)


@hooks.Filters.APP_PUBLIC_HOSTS.add()
def _print_superset_host(hosts: list[str], context_name: t.Literal["local", "dev"]):
    if context_name == "dev":
        hosts.append("{{ CAIRN_HOST }}:2247")
    else:
        hosts.append("{{ CAIRN_HOST }}")
    return hosts


@click.command(
    name="cairn-createuser", help="Create a Cairn user, both in Clickhouse and Superset"
)
@click.option(
    "--bootstrap-dashboards",
    is_flag=True,
    help="Load the default Cairn dashboards to the user's Superset account",
)
@click.option("--admin", is_flag=True)
@click.option(
    "-p",
    "--password",
    help="Specify password from the command line. If undefined, no password will be set. (Ignored with SSO)",
    hide_input=True,
)
@click.option(
    "-c",
    "--course-id",
    "course_ids",
    help="Limit access to a selection of courses (Ignored with SSO).",
    multiple=True,
    hide_input=True,
)
@click.argument("username")
@click.argument("email")
def create_user_command(
    bootstrap_dashboards: bool, admin: bool, password: str, course_ids: list[str], username: str, email: str
) -> t.Iterable[tuple[str, str]]:
    admin_opt = " --admin" if admin else ""

    # TODO can we now simplify the clickhouse image?
    # - get rid of the cairn utility
    # - remove the auth.json file

    create_superset_user = "python ./superset/cairn/ctl.py createuser"
    if password:
        create_superset_user += f" --password={shlex.quote(password)}"
    for course_id in course_ids:
        create_superset_user += f" --course-id={course_id}"
    create_superset_user += f" {admin_opt} {username} {email}"
    yield ("cairn-superset", create_superset_user)

    # Bootstrap dashboards
    if bootstrap_dashboards:
        yield (
            "cairn-superset",
            f"python ./superset/cairn/ctl.py bootstrap-dashboards {username} /app/bootstrap/courseoverview.json",
        )


hooks.Filters.CLI_DO_COMMANDS.add_item(create_user_command)

####### Boilerplate code
# Add the "templates" folder as a template root
hooks.Filters.ENV_TEMPLATE_ROOTS.add_item(
    pkg_resources.resource_filename("tutorcairn", "templates")
)
# Render the "build" and "apps" folders
hooks.Filters.ENV_TEMPLATE_TARGETS.add_items(
    [
        ("cairn/build", "plugins"),
        ("cairn/apps", "plugins"),
    ],
)
# Load patches from files
for path in glob(
    os.path.join(pkg_resources.resource_filename("tutorcairn", "patches"), "*")
):
    with open(path, encoding="utf-8") as patch_file:
        hooks.Filters.ENV_PATCHES.add_item((os.path.basename(path), patch_file.read()))



#! /usr/bin/env python3

import argparse
import json
from time import time

from superset.app import create_app

app = create_app()
app.app_context().push()

from superset.connectors.sqla.models import SqlaTable
from superset.models.core import Database
from superset.models.slice import Slice
from superset.extensions import db, security_manager
import superset.dashboards.commands.importers.v0 as importers
from werkzeug.security import generate_password_hash

# Our convenient library
from superset.cairn import bootstrap as cairn_bootstrap


now = time()


def main():
    parser = argparse.ArgumentParser(
        description="Bootstrap user creation and dashboards"
    )
    subparsers = parser.add_subparsers()

    # Create user
    parser_user = subparsers.add_parser("createuser", help="Create or update user")
    parser_user.add_argument(
        "-d",
        "--db",
        "--database",
        help=(
            "Name of the Superset database to which the user should be granted access."
            " Defaults to the username."
        ),
    )
    parser_user.add_argument(
        "--admin",
        action="store_true",
        help=("Make the user an administrator."),
    )
    parser_user.add_argument(
        "-p",
        "--password",
        help="User password.",
    )
    parser_user.add_argument(
        "--firstname", default="", help="User first name (optional)."
    )
    parser_user.add_argument(
        "--lastname", default="", help="User last name (optional)."
    )
    parser_user.add_argument(
        "-c",
        "--course-id",
        action="append",
        help="Restrict user to access data only from these courses.",
    )
    parser_user.add_argument("username")
    parser_user.add_argument("email")
    parser_user.set_defaults(func=bootstrap_user)

    # Bootstrap dashboards
    parser_dashboards = subparsers.add_parser(
        "bootstrap-dashboards",
        help="Bootstrap datasets and dashboards for a given user",
    )
    parser_dashboards.add_argument(
        "-d",
        "--db",
        "--database",
        help=(
            "Name of the Superset database to which the objects should be linked."
            " By default, this will be the 'openedx-<username>'."
        ),
    )
    parser_dashboards.add_argument("username")
    parser_dashboards.add_argument("path", nargs="+")
    parser_dashboards.set_defaults(func=bootstrap_dashboards)

    args = parser.parse_args()
    args.func(args)


# Note: we'd like to get rid of this command by relying on `superset fab create-user`
# but the "create-user" command fails if the user already exists.
def bootstrap_user(args):
    # Bootstrap database
    database_name = args.db or f"openedx-{args.username}"
    cairn_bootstrap.create_superset_db(args.username, database_name)

    # Get or create user
    user = security_manager.find_user(args.username)
    if user:
        print(f"User '{args.username}' already exists. Skipping creation.")
    else:
        print(f"Creating user '{args.username}'...")
        user = security_manager.add_user(
            args.username,
            args.firstname,
            args.lastname,
            args.email,
            security_manager.find_role("Gamma"),
        )
        if user is None or user is False:
            # This may happen for instance when the email address is already associated
            # to a different username
            raise RuntimeError(
                f"Failed to create user '{args.username}' email='{args.email}'"
            )

    # Set password
    if args.password:
        print("Setting user password...")
        user.password = generate_password_hash(args.password)
        db.session.add(user)
        db.session.commit()

    # Create user role, clickhouse db, etc.
    cairn_bootstrap.setup_user(args.username, course_ids=args.course_id)

    # Associate user to roles
    user_roles = cairn_bootstrap.get_role_names(args.username)
    if args.admin:
        user_roles.append("Admin")
    for role_name in user_roles:
        role = security_manager.find_role(role_name)
        if role in user.roles:
            print(f"Role '{role_name}' is already associated to user.")
        else:
            print(f"Associating role '{role_name}' to user...")
            user.roles.append(role)
            db.session.add(user)
            db.session.commit()

    print("Done.")


# Note: we would like to start using superset's native export/import-dashboards command
# but we failed to get it to work.
def bootstrap_dashboards(args):
    database_name = args.db or args.username
    database = load_database(database_name)
    user = load_user(args.username)

    for path in args.path:
        print(
            f"importing dashboard {path} for user='{user.username}' db='{database.database_name}'... "
        )
        dashboard = load_dashboard_file(path)
        import_dashboard(dashboard, user, database)


def load_database(database_name):
    return (
        db.session.query(Database).filter(Database.database_name == database_name).one()
    )


def load_user(username):
    return (
        db.session.query(security_manager.user_model)
        .filter(security_manager.user_model.username == username)
        .one()
    )


def load_dashboard_file(path):
    with open(path, encoding="utf-8") as f:
        return json.load(f, object_hook=importers.decode_dashboards)


def import_dashboard(data, user, database):
    # Load datasets
    for dataset in data["datasources"]:
        dataset.params = "{}"
        # This should overwrite the existing dataset, if any with the same name
        new_dataset_id = importers.import_dataset(dataset, database.id, now)
        new_dataset = db.session.query(SqlaTable).get(new_dataset_id)
        # Make current user the new owner
        new_dataset.owners.append(user)
        db.session.add(new_dataset)

    for dashboard in data["dashboards"]:
        # Load slices
        new_slices = []
        slices = dashboard.slices[:]
        old_to_new_slice_id_map = {}

        for dashboard_slice in slices:
            # Make sure the new slice does not point to the old slice
            params_dict = dashboard_slice.params_dict
            params_dict["database_name"] = database.name
            params_dict.pop("remote_id")
            dashboard_slice.params = json.dumps(params_dict)
            old_slice_id = dashboard_slice.id
            dashboard_slice.id = None
            # Create new slice
            new_slice_id = importers.import_chart(dashboard_slice, None, now)
            new_slice = db.session.query(Slice).get(new_slice_id)
            old_to_new_slice_id_map[old_slice_id] = new_slice_id
            # Make current user the new owner
            new_slice.owners.append(user)
            new_slices.append(new_slice)
            db.session.add(new_slice)

        # Add slices to dashboard
        dashboard.slices = new_slices
        # Set dashboard owner
        dashboard.owners.append(user)
        # Update position JSON
        position = dashboard.position.copy()
        for chart in position.values():
            if (
                "meta" in chart
                and isinstance(chart["meta"], dict)
                and "chartId" in chart["meta"]
            ):
                chart["meta"]["chartId"] = old_to_new_slice_id_map[
                    chart["meta"]["chartId"]
                ]
        dashboard.position_json = json.dumps(position)

        # Update filter mapping
        metadata = json.loads(dashboard.json_metadata)
        new_filter_scopes = {}
        for old_slice_id, filter_scope in metadata["filter_scopes"].items():
            new_filter_scope = {}
            for filter_name, properties in filter_scope.items():
                properties["immune"] = [
                    old_to_new_slice_id_map[old_immune_id]
                    for old_immune_id in properties["immune"]
                ]
                new_filter_scope[filter_name] = properties
            # Note that filter scope keys are str, not int
            new_filter_scopes[
                str(old_to_new_slice_id_map[int(old_slice_id)])
            ] = new_filter_scope
        metadata["filter_scopes"] = new_filter_scopes
        dashboard.json_metadata = json.dumps(metadata)

        # Load dashboard
        db.session.add(dashboard)

        # Commit changes
        db.session.commit()


if __name__ == "__main__":
    main()



import argparse
import json
import os

import requests
# The escape_string was moved to a submodule. It is suggested to use
# `connection.escape_string` instead, but here we don't have a connection object.
# https://mysqlclient.readthedocs.io/user_guide.html#mysql-c-api-function-mapping
from MySQLdb._mysql import escape_string as sql_escape_string

import lms.startup

lms.startup.run()

from lms.djangoapps.courseware.courses import get_course
from xmodule.modulestore.django import modulestore

with open(os.path.join(os.path.dirname(__file__), "..", "clickhouse-auth.json")) as f:
    CLICKHOUSE_AUTH = json.load(f)


def main():
    parser = argparse.ArgumentParser(
        description="Import course block information into the datalake"
    )
    parser.add_argument(
        "-c", "--course-id", action="append", help="Limit import to these courses"
    )
    args = parser.parse_args()

    module_store = modulestore()
    course_ids = args.course_id or []
    for course in module_store.get_courses():
        if str(course.id) in course_ids or not course_ids:
            import_course(course.id)


def import_course(course_key):
    course_id = str(course_key)
    # Reload course to fetch all children items
    course = get_course(course_key, depth=None)
    print("======================", course_id, course.display_name)
    values = [
        sql_format(
            "('{}', '{}', '{}', '{}', '{}', '{}', '{}')",
            course_id,
            str(child.location),
            child.location.block_id,
            str(position),
            child.display_name or "N/A",
            full_name,
            "true" if child.graded else "false",
        )
        for position, (child, full_name) in enumerate(iter_course_blocks(course))
    ]
    if values:
        print(
            f"Inserting {len(values)} items in course_blocks for course '{course_id}'..."
        )
        make_query(
            sql_format(
                "ALTER TABLE course_blocks DELETE WHERE course_id = '{}';",
                course_id,
            ),
        )
        insert_query = sql_format(
            "INSERT INTO course_blocks (course_id, block_key, block_id, position, display_name, full_name, graded) VALUES "
        )
        insert_query += ", ".join(values)
        make_query(insert_query)


def iter_course_blocks(item, prefix=""):
    prefix += item.display_name or "N/A"
    yield item, prefix
    prefix += " > "
    for child in item.get_children():
        yield from iter_course_blocks(child, prefix=prefix)


def sql_format(template, *args, **kwargs):
    args = [sql_escape_string(arg).decode() for arg in args]
    kwargs = {key: sql_escape_string(value).decode() for key, value in kwargs.items()}
    return template.format(*args, **kwargs)


def make_query(query):
    clickhouse_uri = (
        f"{CLICKHOUSE_AUTH['http_scheme']}://{CLICKHOUSE_AUTH['username']}:{CLICKHOUSE_AUTH['password']}@"
        f"{CLICKHOUSE_AUTH['host']}:{CLICKHOUSE_AUTH['http_port']}/?database={CLICKHOUSE_AUTH['database']}"
    )
    response = requests.post(clickhouse_uri, data=query.encode("utf8"))
    if response.status_code != 200:
        print(response.content.decode())
        raise ValueError("An error occurred while attempting to post a query")


if __name__ == "__main__":
    main()
